# auxfuncs.py

__all__ = [ 'getInput', 'getLetter', 'askYN',
        'measureMinIndent', 'indent', 'alignToMinIndent'
        ]

import sys

def getInput(prompt, default=None, noStrip=False):
    ''' Displays prompt, returns keyboard input as string
    
    Handles Python 2/3 differences, allows default value for blank <Enter>
    '''

    if sys.version_info[0] < 3:
        resp = raw_input(prompt)
    else:
        resp = input(prompt)
    if not noStrip:
        resp = resp.strip()

    if not resp: # eg. blank keyboard input
        if not (default is None):
            resp = default

    return resp

def getLetter(choices, default=None, prompt='', rePrompt=None):
    ''' Returns first letter of keyboard input.
    
    Letter is validated by presence in 'choices' string
    If default is set, <nothing> <Enter> allowed; will return 'default' string.
    '''
    if rePrompt is None:
        rePrompt = prompt
    while True:
        resp = str( getInput(prompt) ).lower()
        if not resp:
            if not ( default is None ):
                return default.lower()
            prompt = rePrompt
            continue
        if resp[0] in choices:
            return resp[0]
        prompt = rePrompt

    # should never get here
    return

def askYN(prompt='', promptSuffix=None, rePrompt=None, default=None,
        returnBoolean=True, allowQuit=False):
    ''' Ask a yes/no type question.

    Can return boolean simplifying code in caller. Since a 'q'uit option
    is often useful, and bool can't handle 3 values, returnBoolean=False
    will return single letter (lowercase) instead.
    
    default=<letter> will set default to a letter as needed
    promptSuffix and rePrompt can be set to a value, or to '' to deactivate

    allowQt parameter refers to the 'Qt' toolkit (unrelated to 'Quit')
    '''

    if promptSuffix is None:
        if allowQuit:
            promptSuffix = ' (y/n/q): '
        else:
            promptSuffix = ' (y/n): '
    if rePrompt is None:
        if allowQuit:
            rePrompt='Enter y/n/q: '
        else:
            rePrompt='Enter y or n: '
    prompt += promptSuffix

    if allowQuit:
        choices = 'ynq'
        returnBoolean = False
    else:
        choices = 'yn'

    resp = getLetter(choices=choices, default=default, prompt=prompt,
            rePrompt=rePrompt)

    if returnBoolean:
        return resp == 'y'
    else:
        return resp
    
def measureMinIndent(stringOrList):
    inputIsList = type(stringOrList) == list
    if inputIsList:
        lineList = stringOrList
    else:
        lineList = stringOrList.splitlines()

    justRealBig = 999
    minIndent = justRealBig
    for thisLine in lineList:
        if not thisLine:  # blank line (just \n)
            continue
        lstripped = thisLine.lstrip()
        if lstripped:
            indent = thisLine.find(lstripped[0])
            if indent < minIndent:
                minIndent = indent
    if minIndent == justRealBig:
        minIndent = 0
    return minIndent

def indent(inputString, shiftCols):
    if not (type(shiftCols) == int):
        print('indent: shiftCols not int')
        return None
    if shiftCols == 0:
        return inputString

    shiftRight = True
    if shiftCols < 0:
        shiftRight = False
        shiftCols = - shiftCols

    shiftList = []
    for thisLine in inputString.splitlines():
        if shiftRight:
            shiftList += [ (' ' * shiftCols) + thisLine ]
        else:
            shiftList += [ thisLine[shiftCols:] ]

    return '\n'.join(shiftList)

def alignToMinIndent(stringOrList, minIndent):
    inputIsList = type(stringOrList) == list
    if inputIsList:
        inputString = '\n'.join(stringOrList)
    else:
        inputString = stringOrList

    currentMin = measureMinIndent(inputString)
    shift = minIndent - currentMin
    if shift:
        inputString = indent(inputString, shift)

    if inputIsList:
        stringOrList = inputString.splitlines()
    else:
        stringOrList = inputString
    return stringOrList
