# indenttagfile.py

__all__ = [ 'itf', 'itftran', 'itftranprint' ]

# This script translates 'indented tag files' into html or other protocols,
# using an internal dictionary which can be supplemented using one or more
# external files. See bottom of this file for internal dictionary.
#
# The hierarchy of tags is controlled by indentation, like blocks of Python.
# Lines starting with '#' are comments.  Example below:
#
# [HTML]
#   [HTMLHEAD]
#     [MINICSSHEAD]
#   [HTMLBODY]
#     [CONTAINER]
#       [ROW]
#          &nbsp;
#       [CARD] large
#         lorem ipsum dolor and so on
#

### GLOBALS #############################################
global doRecursion, tagRE, leaveTagComment, tagNameMaxLen
#########################################################

# a multipass approach without recursion works better, as it turns out
doRecursion = False

leaveTagComment=True
tagNameMaxLen = 30

import pdb, sys, re, os, shlex
from ast import literal_eval

countRange = '{0,' + str(tagNameMaxLen - 1) + '}'

# tag ->[tagName]<- where tagName starts with letter, up to tagNameMaxLen chars
tagREstring = '.*\[[a-zA-Z][^]]{}\].*'.format(countRange)
# literal [ ______^  ^      ^  ^
#    any letter _____|      |  |
#       any char except ] __|  |
#                              |__ count range (between 0 and tagNameMaxLen - 1

tagRE = re.compile(tagREstring)

from auxfuncs import askYN, measureMinIndent, alignToMinIndent
# from .stablecore import askYN
# from .texttools import measureMinIndent, alignToMinIndent

def callerName(callerFrame=None):
    if not callerFrame:
        callerFrame = sys._getframe(1)

    funcName = callerFrame.f_code.co_name
    if not funcName:
        print('callerFrame.f_code.co_name was null')
        return None
    elif funcName == '<module>':
        print('Calling from interpreter?')
        print("callerFrame.f_code.co_name == '<module>'")
        return None

    funcName += '()'

    return funcName

def errorMsg(msg, callerFrame=None):
    if not callerFrame:
        callerFrame = sys._getframe(1)

    whoCalledMe = callerName(callerFrame=callerFrame)
    print('{}: {}'.format(whoCalledMe, msg) )
    return

class Stop(Exception):
    def __init__(self):
        print('Type => del sys.tracebacklimit')
        print('  to show usual traceback on next error (for *other* scripts)')
        # NOTE guaranteed to hit itf.__init__ when using this script,
        # and it can take care of 'del sys.tracebacklimit' there
        # (uncomment if needed)
        sys.tracebacklimit = 0

class itf(object):
    ''' an itf instance represents an indented tag file
    expected to initially have only a 'source' version, which can
    be expanded/translated by having embedded tags replaced with
    their definitions, found in other files
    '''
    def __init__(self, filePath, translatedFilePath=None):
        # Following resets, if needed, will only be needed if using 'Stop'
        # exception while debugging looping issues
        if hasattr(sys, 'tracebacklimit'):
            del sys.tracebacklimit
        #
        if not os.path.isfile(filePath):
            print("File '{}' not found.".format(filePath))
            raise Stop()

        if not (filePath.endswith('.itf')):
            print('Recommend indented tag file name ends with .itf')
            print('My .vimrc has vim expand tabs using spaces for .itf files,')
            print('Which will prevent indentation problems.')

        self.filePath = filePath
        self.contentStringList = []
        self.translatedFilePath = translatedFilePath
        return

    def setContentStringList(self):
        ''' sets self.contentStringList to contents of self.filePath '''
        with open(self.filePath, 'r') as inputFile:
            fileString = inputFile.read()

        tabsFound = False
        self.contentStringList = []
        for line in fileString.splitlines():
            if line and line[0] == '#':  # comment line
                continue
            if '\t' in line:
                tabsFound = True
            self.contentStringList.append(line)

        if tabsFound:
            print('''
            Warning: tab characters found (outside of comment lines)
            Mix of tabs and spaces can muddy indentation.
            If using 'vi' suggest -> :set expandtab
            ''')
        return

    def translate(self, dictFilePathOrPathList=None, useInternalDict=True,
            protocol=None, noPrint=None, toFile=True):
        if noPrint is None:
            noPrint = toFile

        if not self.contentStringList:
            self.setContentStringList()

        isHtml = self.contentStringList[0].startswith('[HTML')

        if toFile and ( not self.translatedFilePath ):
            outputPath = self.filePath
            if outputPath.endswith('.itf'):
                outputPath = outputPath.replace('.itf','')
            if isHtml:
                outputPath += '.html'
            else:
                outputPath += '.xlate'
            if not askYN('Write translation to {}?'.format(outputPath)):
                print('Cancelling.')
                return None
            self.translatedFilePath = outputPath

        translateDict = buildDict(dictFilePathOrPathList,
                useInternalDict=useInternalDict)
        if not translateDict:
            return None

        xlateStringList = translateStringList(self.contentStringList,
            translateDict, protocol=protocol)
        if not xlateStringList:
            print('Problem translating...')
            return None

        xlateString = '\n'.join(xlateStringList)
        xlateString = fixBraces(xlateString)

        if not noPrint:
            print(xlateString)

        if toFile:
            with open(self.translatedFilePath, 'w') as xf:
                xf.write(xlateString)
            return self.translatedFilePath
        else:
            return

def fixBraces(inString):
    # curly braces in templates that *aren't* meant for substitution
    # are *doubled* in dictionary file
    # so here we switch them back to single braces so original
    # html etc. is preserved unbroken
    if '{' in inString:
        inString = inString.replace('{{','{').replace('}}','}')
    return inString

def hasTags(inStringList):
    global tagRE

    hasTags = False
    for line in inStringList:
        if '<!-' in line:
            # skip tags already translated with tag id now enclosed in comment
            line = re.sub('<!-.*->', '', line)
        if tagRE.match(line):
            hasTags = True
            break
    return hasTags

def checkForPseudoTags(inStringList, dict, protocol='html'):
    itfString = '\n'.join(inStringList)

    # if protocol == 'html':
    if True: # for now, just always do this
        itfString = re.sub('<!-[^>]*>', '', itfString)
    tagList = re.findall('\[[^]]*]', itfString)
    
    missingTags = []
    for thisTag in tagList:
        if not (thisTag in dict):
            missingTags += [ thisTag ]

    if missingTags:
        print('Following tag(s) were not found in translation dictionary')
        for thisTag in missingTags:
            sys.stdout.write(thisTag + ' ')  # print w/o newline
        print('')
    return

def translateStringList(srcStringList, translateDict=None, protocol=None):

    sanityCheckAtLoopCount = 100

    xlateStringList = srcStringList

    loopCount = 0
    while hasTags(xlateStringList):

        (startRowIdx, endRowIdx) = findFirstTag(xlateStringList, translateDict)

        if (startRowIdx is None) or (endRowIdx is None):
            checkForPseudoTags(xlateStringList, translateDict, protocol)
            break

        tagExtract = xlateStringList[startRowIdx : endRowIdx]

        #  *if* doRecursion = True, it may happen in next call
        # since expandTag may call *this* func.
        # The only way out is for this loop to end.
        tagXlate = expandTag(tagExtract, translateDict, protocol=protocol)

        if not tagXlate:
            tag = xlateStringList[startRowIdx].split()[0]
            errorMsg("Couldn't expand tag '{}'".format(tag) )
            if not askYN('Continue?'):
                raise Stop()
            print('Will leave untranslated and continue.')
            tagXlate = tagExtract

        xlateStringList = (
                xlateStringList[0 : startRowIdx] + tagXlate +
                xlateStringList[endRowIdx : ]
        )

        loopCount += 1
        if loopCount == sanityCheckAtLoopCount:
            print('Unless you have {}+ [TAG]s in your script, somethings wrong.'
                .format(sanityCheckAtLoopCount) )
            if not askYN('Continue?'):
                raise Stop()
            print("You might change 'sanityCheckAtLoopCount' value in code.")

    return xlateStringList

def expandTag(inStringList, tagDict, protocol=None):
    global doRecursion, leaveTagComment

    if protocol is None:
        protocol = 'html'

    tagNameLine = inStringList[0]
    contentLines = inStringList[1:] # this might be empty

    # following effectively shifts content lines after tag in source file
    # so that the leftmost indent will end up two columns to the right
    # of the position of the tag itself in the translation
    contentLines = alignToMinIndent(contentLines, 2)
    
    tokenList = shlex.split(tagNameLine)  # shlex version of split leaves
    #                                       quoted word groups together
    tagName = tokenList[0]
    parmList = tokenList[1:] # if no parms, empty (won't fail)

    baseIndent = tagNameLine.find( tagName[0] ) # get column of first non-blank
    #                                             char, (first char of tag)

    if tagName in tagDict:
        tagDictValue = tagDict[tagName]
    else:
        errorMsg("Tag '{}' not found in dictionary.".format(tagName) )
        return None

    tagCustomPython = None
    if type(tagDictValue) == list:
        tagTemplate = tagDictValue[0]
        if len(tagDictValue) > 1:
            tagCustomPython = tagDictValue[1]
    else:
        tagTemplate = tagDictValue

    if doRecursion:
        # call my caller
        contentLines = translateStringList(contentLines, tagDict,
                protocol=protocol)
        # above translates 'inner' content of this tag *before* generating
        # a translation for the tag itself.
        # If doRecursion is false, inner content will be translated in
        # the next loop iteration of translateStringList()
        # *after* we return from this function

    # turn line list into single string for following translation
    contentString = '\n'.join(contentLines)

    ######################################################################
    # THIS IS WHERE ACTUAL TRANSLATION HAPPENS

    if tagCustomPython:
        xlateString = None
        _locals = locals()

        try:
            exec(tagCustomPython, globals(), _locals)
        except:
            print('Error executing custom Python tag.')
            if askYN('View code?'):
                print(tagCustomPython)
            if not askYN('Continue?'):
                raise # re-raise previous exception to allow pdb.pm()

        if not ( 'xlateString' in _locals and _locals['xlateString'] ):
            errorMsg(" 'tagCustomPython' code didn't set xlateString ")
            return None
        xlateString = _locals['xlateString']
    else:
        if not containsSubstitutionBraces(tagTemplate):
            # just copy the template as our translation
            xlateString = tagTemplate
        else:
            # {content} in tagTemplate will be replaced by contentString,
            # and {0}, {1} ... will be replaced by the parameters
            # following [<tagName>] (on same line) in source file
            try:
                xlateString = tagTemplate.format( *parmList,
                        content=contentString )
            except IndexError:
                print("Missing parameter for '{}' tagTemplate".format(tagName))
                return None

    ######################################################################

    # turn string back into a list of lines
    xlateStringList = xlateString.splitlines()

    minIndent = measureMinIndent(xlateStringList)

    # to align the translation nicely, we take the column position of 
    # a tag (baseIndent), and then shift the tag content to the right,
    # to be lined up with tag position in original file ** + 2 **
    shift = baseIndent - minIndent + 2

    if shift > 0:   # to shift right, we'll precede with spaces
        indentString = ' ' * shift
        colIdx = 0
    else:           # to shift left, we'll index into the string
        indentString = ''
        colIdx = - shift

    for rowIdx, thisLine in enumerate(xlateStringList):
        xlateStringList[rowIdx] = indentString + thisLine[colIdx:]

    if leaveTagComment:
        comment = ''
        if protocol == 'html':
            comment = '<!-- {} -->'.format(tagNameLine.strip())

        # may handle commenting for other protocols later
        # for now, if not html, we just won't include commented tag

        # also, don't comment [HTML] tag, since we want its translation,
        # <!DOCTYPE html> to occur on *first* line of output
        if comment and (not ('[HTML]' in tagNameLine)):
            comment = (' ' * baseIndent) + comment
            xlateStringList = [ comment ] + xlateStringList
            
    return xlateStringList

def buildDictFromSetStrings(setList, allowedVarsList): 
    thisDict = {}

    for setStmt in setList:
        if not '=' in setStmt:
            print("'{}' not assignment statement -- skipping.".format(setStmt))
            continue

        (varName, value) = nameAndValueFromAssignString(setStmt)

        if varName in allowedVarsList:
            thisDict[varName] = literal_eval(value)
        else:
            print("'{}' not in allowedVarsList".format(varName))

    return thisDict

def nameAndValueFromAssignString(assignString):
    if not ('=' in assignString):
        print("No '=' found in string")
        return None

    assignString = assignString.replace(' =','=')
    assignString = assignString.replace('= ','=')
    return assignString.split('=')

def containsSubstitutionBraces(formatString):
    
    # turn (possible) multi-line string into one long string
    # so that regular expression below will work...
    formatString = formatString.replace('\n','')

    hasSubBraces = False
    if '{' in formatString:
        if ('{content}' in formatString) or ('{}' in formatString):
            hasSubBraces = True
        else:
            numberedBraceRE = re.compile('.*\{[0-9]\}.*')
            if numberedBraceRE.match(formatString):
                hasSubBraces = True
    return hasSubBraces

def findFirstTag(inStringList, tagDict):
    '''
    Looks in inStringList for first 'tag'
    and returns index to first *and* last line of tag content
    A tag can potentially be a single line, or have several lines of content.

    The end of tag content is either the end of a file, or a non blank line
    with indent to *left* of the tag position (just as the end of a block
    is defined in Python)
    '''

    if not inStringList:
        return (None, None)

    tagList = [ key for key in tagDict ]

    firstTagIndentColumn = None
    startRowIdx = None
    endRowIdx = None
    brokeOut = False
    for rowIdx, thisLine in enumerate(inStringList):
        lstripped = thisLine.lstrip()
        if not lstripped: # blank line
            continue

        nonBlankIndentColumn = thisLine.find(lstripped[0])

        if not (firstTagIndentColumn is None):  # has been set
            if nonBlankIndentColumn <= firstTagIndentColumn:
                ############### have reached end of block (higher level indent)
                brokeOut = True
                break
                ###############

        else:  #  firstTagIndentColumn **UNset**
            tagIndentColumn = None
            #                                ___________
            for tryMatch in tagList: # ----/           \
                if tryMatch in thisLine: #              |
                    if '<!-' in thisLine: #              |
                        # html comment, likely as result |
                        # of this tag already translated |
                        # but left in commented form     |
                        continue # ---------------------/
                    tagIndentColumn = thisLine.find(tryMatch)
                    break
            #  ____/
            # /
            if not (tagIndentColumn is None):  # is set, AND still true if zero
                firstTagIndentColumn = tagIndentColumn
                startRowIdx = rowIdx

    if not brokeOut:  # scanned to end of loop without finding another tag
        rowIdx += 1   # so add 1, because final rowIdx value
        #               points to line *within* this tag

    endRowIdx = rowIdx

    return (startRowIdx, endRowIdx)

def checkDict(theDict):
    global tagNameMaxLen
    for key in theDict:
        if len(key) > (tagNameMaxLen + 2):
            print("Problem: tag name '{}' is too long.".format(key))
            print('tagNameMaxLen is set to {}'.format(tagNameMaxLen))
    return

def buildDict(dictFilePathOrPathList=None, useInternalDict=True):
    if useInternalDict:
        resultDict = internalDict
        if not dictFilePathOrPathList:
            checkDict(resultDict)
            return resultDict
    else:
        if not dictFilePathOrPathList:
            errorMsg(
                'If useInternalDict=False, must supply dictFilePathOrPathList')
            return None
        resultDict = {}

    if type(dictFilePathOrPathList) == list:
        dictFilePathList = dictFilePathOrPathList
    else:
        dictFilePathList = [ dictFilePathOrPathList ]

    for thisDictFile in dictFilePathList:
        if not os.path.isfile(thisDictFile):
            errorMsg("Dict file '{}' not found.".format(thisDictFile))
            continue

        with open(thisDictFile, 'r') as df:
            dictInitString = df.read()

        dictInitString = 'workDict = {\n' + dictInitString + '\n}'
        exec(dictInitString)
        exec('resultDict.update(workDict)')

    checkDict(resultDict)

    return resultDict

revealMiniDir = os.path.expanduser('~/Documents/webpages/revealAndMini')

def itftran(sourceFile, outputFile=None, dictFile=None):
    thisitf = itf(sourceFile, translatedFilePath=outputFile)
    xlateFile = thisitf.translate(dictFile, useInternalDict=True)
    return

def itftranprint(sourceFile, dictFile=None):
    thisitf = itf(sourceFile)
    xlateFile = thisitf.translate(dictFile, useInternalDict=True, toFile=False,
            noPrint=False)
    return

internalDict = {

# *** Dictionary entries follow these comments.
#     This program can also load such entries from external file(s).
#
#     The actual 'dict = {' (above) and closing brace
#     should *not* be included in external files.
#     It will be added by my code before 'exec'uting.
#     This allows the code to control the name of the dictionary.
#
#########################################################################
#
# Note most common format of entries:
#
#   Tag name
#     v
# '[EXAMPLE]': '''      ___
# <div class="example">    \_______ multi-line 'template'
# {content}                |
# </div>    ______________/
# '''[1:-1], <---- always include comma
#      ^           (it won't 'break' dict to have one after last entry)
#      |
#      The [1:-1] strips off newlines from start and end of '''string'''
#      Having these newlines simplifies checking proper indentation/alignment
#      in template compared to:
#
#      '''<div class="example">
#      {content}
#      </div>'''
#
# -----------------------------------------------------------------------
#
# CURLY BRACES ARE DOUBLED {{  }} in templates below if they are
# not meant to have values substituted into them -- by my code, using format()
# After translation is complete, they will be converted back to single braces.
#
# -----------------------------------------------------------------------
#
# The 'value' for each dictionary 'key' can be either a string, or a list.
# The string (or first element of list) is the 'template' to use for the
# 'tag' (key).
# If dict value *is* list, then *second* ( [1] ) element can be a string
# of Python code to execute to translate the template (and tag arguments)
# to the 'result' string
#
# -----------------------------------------------------------------------
#
# If Python code is *not* included, my code does following:
#
#     xlateString = tagTemplate.format( *parmList, content=contentString )
#
# where parmList is space delimited list of parameters on same line as
# '[tagName]' in the source file, and contentString is content
# included on lines *following* line with tag, until end of tag block
# (determined by indent levels)
#
# Exception to above: if template does *not* have any {content} or {0},
# {1}, etc, then format() will *not* be used, template will simply be
# copied to translation output.
#
#########################################################################

'[HTML]': '''
<!DOCTYPE html>
<!--
The following html was generated from a template using a Python application
written by David Magill.

Thanks to Angelos Chalaris for developing the excellent css library 'minicss'
See documentation at https://minicss.org/
-->
<html>
{content}
</html>
'''[1:-1],

'[HTMLHEAD]': '''
<head>
{content}
</head>
'''[1:-1],

# use [MINICSSHEAD] within [HTMLHEAD] as needed
# tag requires style parm eg: [STYLE.BLUEGRAYSEMITRAN]
'[MINICSSHEAD]': '''
<meta http-equiv="content-type" content="text/html; charset=UTF-8">
<title>index</title>
<link rel="stylesheet" href="http://gis.sent.com/css/mini-default.min.css">
<!-- see minicss.org for documentation -->
<meta name="viewport" content="width=device-width, initial-scale=1">
{0}
'''[1:-1],

# use [REVEALHEAD] within [HTMLHEAD] as needed
'[REVEALHEAD]': '''
<meta charset="utf-8">
<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=no">

<link rel="stylesheet" href="http://gis.sent.com/css/reset.css">
<link rel="stylesheet" href="http://gis.sent.com/css/reveal.css">
<link rel="stylesheet" href="http://gis.sent.com/css/theme/black.css">

<!-- Theme used for syntax highlighting of code -->
<link rel="stylesheet" href="http://gis.sent.com/lib/css/monokai.css">

<!-- Printing and PDF exports -->
<script>
	var link = document.createElement( 'link' );
	link.rel = 'stylesheet';
	link.type = 'text/css';
	link.href = window.location.search.match( /print-pdf/gi ) ? 'css/print/pdf.css' : 'css/print/paper.css';
	document.getElementsByTagName( 'head' )[0].appendChild( link );
</script>
'''[1:-1],

'[HTMLBODY]': '''
<body>
{content}
</body>
'''[1:-1],

# use [REVEALBODY] instead of [HTMLBODY] to include needed script for slides
'[REVEALBODY]': '''
<body>
{content}
<script src="http://gis.sent.com/js/reveal.js"></script>

<script>
	// More info about config & dependencies:
	// - https://github.com/hakimel/reveal.js#configuration
	// - https://github.com/hakimel/reveal.js#dependencies
	Reveal.initialize({{
		hash: true,
		dependencies: [
			{{ src: 'plugin/markdown/marked.js' }},
			{{ src: 'plugin/markdown/markdown.js' }},
			{{ src: 'plugin/highlight/highlight.js' }},
			{{ src: 'plugin/notes/notes.js', async: true }}
		]
	}});
</script>
</body>
'''[1:-1],

'[LINELIST]': [ None,  # no template, just custom Python
'''
xlateString = ''
for _thisLine in contentString.splitlines():
  xlateString += _thisLine + ' <br>\\n' # must double escape newline for exec()
'''[1:-1]
],

'[LINKLIST]': [ None,  # no template, just custom Python
'''
xlateString = ''
try:
    for _thisLine in contentString.splitlines():
      if '](' in _thisLine:
        _textAndUrl = re.findall('\\[[^]]+\\]\\([^)]+\\)', _thisLine)[0]
        _text =       re.findall('\\[[^]]*\\]', _textAndUrl)[0][1:-1]
        _url =        re.findall('\\]\\([^)]*\\)', _textAndUrl)[0][2:-1]
        _linkTag = '<a href="{}" target="_blank">{}</a>'.format(_url, _text)
        _thisLine = _thisLine.replace(_textAndUrl, _linkTag)
      xlateString += _thisLine + ' <br>\\n'
      # must double escape newline for exec()
except:
    print('Problem with content of [LINKLIST] tag')
    raise # re-raise the same exception
'''[1:-1]
],

'[STYLE.BLUEGRAYSEMITRAN]': [ '''
<style>
  body {{
  background-color: #999999;
  --card-back-color: #bbbbcc;
  --card-border-color: #7777aa;
  --a-visited-color: MidnightBlue;
  --a-link-color: MidnightBlue;
  {_bgi}
  opacity: 0.65;
  }}
</style>
'''[1:-1],
''' # Python code for STYLE.BLUEGRAYSEMITRAN
xlateString = ''
_bgi = "background-image: url('images/globeFaded.jpg');"
try:
    xlateString = tagTemplate.format(_bgi = _bgi)
except:
    print('Problem with [STYLE.BLUEGRAYSEMITRAN] tag')
    raise # re-raise same exception
'''[1:-1]
],

'[STYLE.BLGRWITHGLOBE]': '''
<style>
  body {
  background-color: #999999;
  --card-back-color: #bbbbcc;
  --card-border-color: #7777aa;
  --a-visited-color: MidnightBlue;
  --a-link-color: MidnightBlue;
  background-image: url('images/globeFaded.jpg');
  opacity: 1.0;
  }
</style>
'''[1:-1],

# [SLIDEDECK] must precede/enclose slides
'[SLIDEDECK]': '''
<div class="reveal">
<div class="slides">
{content}
</div>
</div>
'''[1:-1],

# use [SLIDE] slideName
'[SLIDE]': '''
<section>{0}
{content}
</section>
'''[1:-1],

'[CONTAINER]': '''
<div class="container">
{content}
</div> <!-- end container -->
'''[1:-1],

'[ROW]': '''
<div class="row">
{content}
</div> <!-- end row -->
'''[1:-1],

# see https://minicss.org/docs#grid for more 'col' options not handled here
# use [COL] sm|md|lg # (where # is 0-12)
# the number specifies a column width that'll be used for specified screen size
# unspecified screen sizes will use column width given for largest screen size
# smaller than the unspecified size
#
'[COL]': '''
<div class="col-{0}-{1}">
{content}
</div> <!-- end col -->
'''[1:-1],

'[COL12-9-7]': '''
<div class="col-sm-12 col-md-9 col-lg-7">
{content}
</div> <!-- end col -->
'''[1:-1],

'[COL7-6-5]': '''
<div class="col-sm-7 col-md-6 col-lg-5">
{content}
</div> <!-- end col -->
'''[1:-1],

# use [CARD] small|large|fluid (small/large is 240/480 px wide)
# small/large/fluid param will end up in {0}
# 'fluid' cards take available space as needed per content
#
'[CARD]': '''
<div class="card {0}">
{content}
</div> <!-- end card -->
'''[1:-1],

# [IMAGE] tag (format below)
#
#   req'd _           _________   ____ both of these are optional
#          V         V           V
# [IMAGE] <filename> width="###" height="###"
'[IMAGE]': [
#          ^ list follows: indicator that tagTemplate
#            is accompanied by a 'tagCustomPython' string

# tagTemplate
'''
<img src="{filename}" alt="alt text" width="{width}" height="{height}"> 
'''[1:-1],

# custom (exec) Python for this tag -- xlateString used by caller
'''
parmDict = { 'width': '500', 'height': '600' } # defaults
tagImgFname = parmList[0]
parmUpdateDict = buildDictFromSetStrings(parmList[1:], ['width', 'height'] )
parmDict.update(parmUpdateDict)

xlateString = tagTemplate.format(filename=tagImgFname,
        width=parmDict['width'], height=parmDict['height'])
'''[1:-1]
    ]

}
